//! This module provides a generic interface to the various structs that Frame generates to realize
//! name bindings.

use std::any::Any;
use std::cell::{Ref, RefCell};
use std::rc::Rc;

/// Environments associate names (i.e. variables/parameters) with values.
///
/// Environments are defined as a trait rather than a simple HashMap so that we can directly reuse
/// the various structs generated by Frame rather than constructing a bunch of parallel data
/// structures each time we query the state of the machine.
///  
/// It is common for a particular environment to be absent because there are no variables of the
/// kind held by that environment. We represent absent environments by an empty environment rather
/// than using an `Option` type because it simplifies the interface and because the distinction
/// between `None` and `Some(empty)` is not significant.
pub trait Environment {
    /// Is this the empty environment?
    fn is_empty(&self) -> bool {
        false
    }
    /// Get the value associated with a name.
    fn lookup(&self, name: &str) -> Option<&dyn Any>;
}

/// The trivial empty enviorment. This can be used in place of an environment when that environment
/// is absent.
#[derive(Clone, Debug, Eq, Hash, PartialEq, PartialOrd, Ord)]
pub struct Empty {}

impl Empty {
    /// Create a new empty environment.
    pub fn new() -> Empty {
        Empty {}
    }

    /// Create a new reference-counted pointer to an empty environment.
    pub fn new_rc() -> Rc<Empty> {
        Rc::new(Empty::new())
    }
}

impl Default for Empty {
    fn default() -> Self {
        Empty::new()
    }
}

impl Environment for Empty {
    fn is_empty(&self) -> bool {
        true
    }
    fn lookup(&self, _name: &str) -> Option<&dyn Any> {
        None
    }
}

impl<'a, T: Environment> Environment for Ref<'a, T> {
    fn is_empty(&self) -> bool {
        (**self).is_empty()
    }
    fn lookup(&self, name: &str) -> Option<&dyn Any> {
        (**self).lookup(name)
    }
}

impl<T: Environment> Environment for RefCell<T> {
    fn is_empty(&self) -> bool {
        self.borrow().is_empty()
    }
    fn lookup(&self, name: &str) -> Option<&dyn Any> {
        unsafe { (*self.as_ptr()).lookup(name) }
    }
}

#[allow(clippy::approx_constant)]
#[cfg(test)]
mod tests {
    use super::*;

    struct TestArgs {
        x: i32,
        y: bool,
        z: Option<f32>,
    }

    impl Environment for TestArgs {
        fn lookup(&self, name: &str) -> Option<&dyn Any> {
            match name {
                "x" => Some(&self.x),
                "y" => Some(&self.y),
                "z" => Some(&self.z),
                _ => None,
            }
        }
    }

    #[test]
    fn empty_environment_is_empty() {
        let empty = Empty::new();
        assert!(empty.is_empty());
    }

    #[test]
    fn empty_environment_returns_none() {
        let empty = Empty::new();
        assert!(empty.lookup("x").is_none());
        assert!(empty.lookup("y").is_none());
        assert!(empty.lookup("z").is_none());
    }

    #[test]
    fn struct_environment_not_empty() {
        let args = TestArgs {
            x: 42,
            y: false,
            z: Some(3.14),
        };
        assert!(!args.is_empty());
    }

    #[test]
    fn struct_environment_lookup_success() {
        let args = TestArgs {
            x: 42,
            y: false,
            z: Some(3.14),
        };

        let opt_x = args.lookup("x");
        assert!(opt_x.is_some());
        let opt_i32 = opt_x.unwrap().downcast_ref::<i32>();
        assert!(opt_i32.is_some());
        assert_eq!(*opt_i32.unwrap(), 42);

        let opt_y = args.lookup("y");
        assert!(opt_y.is_some());
        let opt_bool = opt_y.unwrap().downcast_ref::<bool>();
        assert!(opt_bool.is_some());
        assert!(!*opt_bool.unwrap());

        let opt_y = args.lookup("z");
        assert!(opt_y.is_some());
        let opt_bool = opt_y.unwrap().downcast_ref::<Option<f32>>();
        assert!(opt_bool.is_some());
        assert_eq!(*opt_bool.unwrap(), Some(3.14));
    }

    #[test]
    fn struct_environment_lookup_type_error() {
        let args = TestArgs {
            x: 42,
            y: false,
            z: Some(3.14),
        };
        let opt_x = args.lookup("x");
        assert!(opt_x.is_some());
        assert!(opt_x.unwrap().downcast_ref::<bool>().is_none());

        let opt_y = args.lookup("y");
        assert!(opt_y.is_some());
        assert!(opt_y.unwrap().downcast_ref::<i32>().is_none());

        let opt_z = args.lookup("z");
        assert!(opt_z.is_some());
        assert!(opt_y.unwrap().downcast_ref::<f32>().is_none());
        assert!(opt_y.unwrap().downcast_ref::<Option<i32>>().is_none());
    }

    #[test]
    fn struct_environment_lookup_undefined() {
        let args = TestArgs {
            x: 42,
            y: false,
            z: Some(3.14),
        };
        assert!(args.lookup("w").is_none());
    }
}
